/************************************************************************************
Blackship RT-Component
Copyright (c) 2010, Segway-Japan, Ltd.
All rights reserved.

Contact us if you use this software for sell.
If you use this software not for sell, you can use this software under BSD lisence.
See the files LICENSE.TXT and LICENSE-BSD.TXT for more details.                     
************************************************************************************/

#include <math.h>
#include "Blackship.h"
#include "bs_if.h"



#if RTM_MAJOR_VERSION >= 1
#else
#include "ace/ACE.h"
#endif


#define USE_SLIP  (1)




// Module specification
// <rtc-template block="module_spec">
static const char* blackship_spec[] =
  {
    "implementation_id", "Blackship",
    "type_name",         "Blackship",
    "description",       "Blackship Component",
    "version",           "1.0.0",
    "vendor",            "SegwayJapan",
    "category",          "Generic",
    "activity_type",     "SPORADIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.Kspeed", "1.0",
    "conf.default.Kturn", "1.6",
    "conf.default.wheelwidth", "0.405",
    "conf.default.wheelradiusL", "0.1535",
    "conf.default.wheelradiusR", "0.1535",
    "conf.default.wheelcount1rot", "2294.6",
    "conf.default.str_port", "/dev/ttyUSB0",
    "conf.default.Kodoxy", "1.0",
    "conf.default.Kodotheta", "1.0",
    "conf.default.Kslip", "0.0",
    "conf.default.mps2bs", "152.0",
    "conf.default.ODOoutput", "0",
    "conf.default.simulator", "0",
    "conf.default.idIIS", "250",
    "conf.default.startxIIS", "0.0",
    "conf.default.startyIIS", "0.0",
    "conf.default.startthetaIIS", "0.0",
    "conf.default.MaxV", "2.0",
    "conf.default.MaxW", "2.0",

    ""
  };
// </rtc-template>

Blackship::Blackship(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_InputIn("Input", m_Input),
    //m_SpeedLIn("SpeedL", m_SpeedL),
    //m_SpeedRIn("SpeedR", m_SpeedR),
    m_VelocityIn("Velocity", m_Velocity),
    m_VelocityIISIn("VelocityIIS", m_VelocityIIS),
    m_Velocity2DIISIn("Velocity2DIIS", m_Velocity2DIIS),
    m_EncoderLOut("EncoderL", m_EncoderL),
    m_EncoderROut("EncoderR", m_EncoderR),
    m_OdometryOut("Odometry", m_Odometry),
    m_OdometryBOut("OdometryB", m_OdometryB),
    m_RobotVelIISOut("RobotVelIIS", m_RobotVelIIS),
    m_RobotPosIISOut("RobotPosIIS", m_RobotPosIIS),
    m_RobotVel2DIISOut("RobotVel2DIIS", m_RobotVel2DIIS),
    m_RobotPose2DIISOut("RobotPose2DIIS", m_RobotPose2DIIS),
    // </rtc-template>
	dummy(0)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("Input", m_InputIn);
  //registerInPort("SpeedL", m_SpeedLIn);
  //registerInPort("SpeedR", m_SpeedRIn);
  registerInPort("Velocity", m_VelocityIn);
  registerInPort("VelocityIIS", m_VelocityIISIn);
  registerInPort("Velocity2DIIS", m_Velocity2DIISIn);
  
  // Set OutPort buffer
  registerOutPort("EncoderL", m_EncoderLOut);
  registerOutPort("EncoderR", m_EncoderROut);
  registerOutPort("Odometry", m_OdometryOut);
  registerOutPort("OdometryB", m_OdometryBOut);
  registerOutPort("RobotVelIIS", m_RobotVelIISOut);
  registerOutPort("RobotPosIIS", m_RobotPosIISOut);
  registerOutPort("RobotVel2DIIS", m_RobotVel2DIISOut);
  registerOutPort("RobotPose2DIIS", m_RobotPose2DIISOut);

  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

}

Blackship::~Blackship()
{
}


RTC::ReturnCode_t Blackship::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("Kspeed", m_Kspeed, "1.0");
  bindParameter("Kturn", m_Kturn, "1.6");
  bindParameter("wheelwidth", m_wheelwidth, "0.405");
  bindParameter("wheelradiusL", m_wheelradiusL, "0.1535");
  bindParameter("wheelradiusR", m_wheelradiusR, "0.1535");
  bindParameter("wheelcount1rot", m_wheelcount1rot, "2294.6");
  bindParameter("str_port", m_str_port, "/dev/ttyUSB0");
  bindParameter("Kodoxy", m_Kodoxy, "1.0");
  bindParameter("Kodotheta", m_Kodotheta, "1.0");
  bindParameter("Kslip", m_Kslip, "0.0");
  bindParameter("mps2bs", m_mps2bs, "152.0");
  bindParameter("ODOoutput", m_ODOoutput, "0");
  bindParameter("simulator", m_simulator, "0");
  bindParameter("idIIS", m_idIIS, "250");
  bindParameter("startxIIS", m_startxIIS, "0.0");
  bindParameter("startyIIS", m_startyIIS, "0.0");
  bindParameter("startthetaIIS", m_startthetaIIS, "0.0");
  bindParameter("MaxV", m_MaxV, "2.0");
  bindParameter("MaxW", m_MaxW, "2.0");
  
  // </rtc-template>


  Speed2[0] = 0;
  Speed2[1] = 0;
  Speed = 0;
  Turn = 0;


  return RTC::RTC_OK;
}



/*
RTC::ReturnCode_t Blackship::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Blackship::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Blackship::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t Blackship::onActivated(RTC::UniqueId ec_id)
{
  PortFd = blackship_open((char*)m_str_port.c_str());

  Status.wheelwidth = m_wheelwidth;
  Status.wheelradiusL = m_wheelradiusL;
  Status.wheelradiusR = m_wheelradiusR;
  Status.count1rot = m_wheelcount1rot;

  Status.odox     = -m_startyIIS;
  Status.odoy     =  m_startxIIS;
  Status.odotheta =  m_startthetaIIS;

  Stop = 0;

  Status.odotime = 
  Status.odotime_prev = coil::gettimeofday();

  return RTC::RTC_OK;
}



RTC::ReturnCode_t Blackship::onDeactivated(RTC::UniqueId ec_id)
{
  blackship_close(PortFd);
  PortFd = 0;

  return RTC::RTC_OK;
}



/*!
  @brief send and recv Blackship command and status
  @param[in]
  @return
*/
void bs_odometry(bs_status& status, double kxy, double ktheta, double kslip, bool ODOoutput)
{
  double pi = 3.1415926535;

  //setting
  double wheelrL = status.wheelradiusL;
  double wheelrR = status.wheelradiusR;
  double radius = status.wheelwidth * 0.5;
  double count1rot = status.count1rot;

  if (status.odofirst) {
    status.int_left_prev  = status.int_left_wheel;
    status.int_right_prev = status.int_right_wheel;
    status.odofirst = false;
  }


  const static double meterpercountL = 2.0 * wheelrL * pi / count1rot;
  const static double meterpercountR = 2.0 * wheelrR * pi / count1rot;
  const static int iterate = 64;

  double countL = (double)(status.int_left_wheel  - status.int_left_prev);
  double countR = (double)(status.int_right_wheel - status.int_right_prev);


  countR /= (double)iterate;
  countL /= (double)iterate;

  double meterR = countR * meterpercountR;
  double meterL = countL * meterpercountL;

#if USE_SLIP
  double diff = meterR - meterL;
  double slip = diff * kslip;
  double meterRslip = meterR - slip;
  double meterLslip = meterL + slip;
  double forward = (meterRslip + meterLslip) * 0.5;
  double diffR = meterRslip - forward;
#else
  double forward = (meterR + meterL) * 0.5;
  double diffR = meterR - forward;
#endif
  double addtheta = diffR / radius;


  status.addx = 0.0;
  status.addy = 0.0;
  status.addtheta = 0.0;
  for (int i = 0; i < iterate; i++) {
    status.odoy += forward * cos(status.odotheta) * kxy;
    status.odox -= forward * sin(status.odotheta) * kxy;
    status.odotheta += addtheta * ktheta;

    status.addy += forward * kxy;
    status.addx += 0.0;
    status.addtheta += addtheta * ktheta;
  }


  if (ODOoutput) {
    printf("#odo %f %f %f enc %d %d   wheelrL %f wheelrR %f radius %f count1rot %f\n",
	   (float)status.odox, (float)status.odoy, (float)status.odotheta,
	   status.int_left_wheel, status.int_right_wheel,
	   wheelrL, wheelrR, radius, count1rot);
  }


  status.int_left_prev  = status.int_left_wheel;
  status.int_right_prev = status.int_right_wheel;
}


RTC::ReturnCode_t Blackship::onExecute(RTC::UniqueId ec_id)
{
  if (PortFd == 0) {
    return RTC::RTC_OK;
  }

  if (m_InputIn.isNew()) {
    while (!m_InputIn.isEmpty()) {
      m_InputIn.read();
    }
    Stop = m_Input.data;
  }



  double mps2bs = m_mps2bs;


  if (m_VelocityIn.isNew() || 
      m_VelocityIISIn.isNew() || 
      m_Velocity2DIISIn.isNew()) {


    double speedmps = 0.0;
    double turnrps = 0.0;

    if (m_VelocityIn.isNew()) {
      while (!m_VelocityIn.isEmpty()) {
	m_VelocityIn.read();
      }
      speedmps = m_Velocity.v;
      turnrps = m_Velocity.w;
    }
    if (m_VelocityIISIn.isNew()) {
      while (!m_VelocityIISIn.isEmpty()) {
	m_VelocityIISIn.read();
      }
      speedmps = m_VelocityIIS.vx;
      turnrps = m_VelocityIIS.w;
    }
    if (m_Velocity2DIISIn.isNew()) {
      while (!m_Velocity2DIISIn.isEmpty()) {
	m_Velocity2DIISIn.read();
      }

      speedmps = m_Velocity2DIIS.data.vx;
      turnrps = m_Velocity2DIIS.data.va;
    }

    if (fabs(speedmps) > m_MaxV) {
      speedmps = (speedmps > 0.0)? m_MaxV : -m_MaxV;
    }
    if (fabs(turnrps) > m_MaxW) {
      turnrps = (turnrps > 0.0)? m_MaxW : -m_MaxW;
    }
 
    Speed = (int) (mps2bs * m_Kspeed * speedmps);
#if UEE_SLIP
    Turn = (int) (mps2bs * m_Kturn * (m_wheelwidth*0.5 * turnrps) / (1.0 - 2.0 * m_Kslip));
#else
    Turn = (int) (mps2bs * m_Kturn * m_wheelwidth*0.5 * turnrps);
#endif
    //printf("mps %f bs %d mps2bs %f\n", (float)speedmps, Speed, mps2bs);
  }


  if (Stop) {
    Speed = 0;
    Turn = 0;
  }


  int s[2];
  s[0] = Speed;
  s[1] = Speed;
  s[0] -= Turn;
  s[1] += Turn;
  s[0] = (s[0] > 100)?   100 : s[0];
  s[0] = (s[0] < -100)? -100 : s[0];
  s[1] = (s[1] > 100)?   100 : s[1];
  s[1] = (s[1] < -100)? -100 : s[1];

  blackship_set_speed(PortFd, s[0], s[1]);



  //-------------------------------
  //get encoder
  //-------------------------------
  int encoder[2];
  Status.odotime_prev = Status.odotime;
  Status.odotime = blackship_get_encoder2(PortFd, encoder[0], encoder[1]);

  Status.int_left_wheel = encoder[0];
  Status.int_right_wheel = encoder[1];
  bs_odometry(Status, m_Kodoxy, m_Kodotheta, m_Kslip, m_ODOoutput);


  //--------------------------------------------------------------------------------
  //output velocity
  //  two wheeled blackship can not run at 1 m/sec. max speed is about 0.4m/sec
  //
  //--------------------------------------------------------------------------------
#if 0
  if (m_simulator == 0 ||
      fabs((float)Status.addx) > 0.0 ||
      fabs((float)Status.addy) > 0.0 || 
      fabs((float)Status.addtheta) > 0.0) {
#endif
   //culc speed
    coil::TimeValue tdiff = Status.odotime - Status.odotime_prev;
    double sec    = (double)((double)tdiff.sec() + (double)tdiff.usec() / 1000000.0);
    double dx     = Status.addx / sec;     //(m/sec)
    double dy     = Status.addy / sec;     //(m/sec)
    double dtheta = Status.addtheta / sec; //(rad/sec)

    //SJ to IIS coord
    double dxIIS = dy;
    double dyIIS = -dx;
    double dthetaIIS = dtheta;

    m_RobotVelIIS.id.length(1);
    m_RobotVelIIS.id[0]   = m_idIIS;
    m_RobotVelIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotVelIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotVelIIS.vx      = dxIIS;
    m_RobotVelIIS.vy      = dyIIS;
    m_RobotVelIIS.w       = dthetaIIS;

    m_RobotVel2DIIS.id.length(1);
    m_RobotVel2DIIS.id[0]   = m_idIIS;
    m_RobotVel2DIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotVel2DIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotVel2DIIS.data.vx = dxIIS;
    m_RobotVel2DIIS.data.vy = dyIIS;
    m_RobotVel2DIIS.data.va = dthetaIIS;
#if 0    
  }
#endif

  {
    //SJ to IIS coord
    m_RobotPosIIS.id.length(1);
    m_RobotPosIIS.id[0]   = m_idIIS;
    m_RobotPosIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotPosIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotPosIIS.x       = Status.odoy;
    m_RobotPosIIS.y       = -Status.odox;
    m_RobotPosIIS.theta   = Status.odotheta;

    m_RobotPose2DIIS.id.length(1);
    m_RobotPose2DIIS.id[0]   = m_idIIS;
    m_RobotPose2DIIS.tm.sec  = (::CORBA::ULong) Status.odotime.sec();
    m_RobotPose2DIIS.tm.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);
    m_RobotPose2DIIS.data.position.x       = Status.odoy;
    m_RobotPose2DIIS.data.position.y       = -Status.odox;
    m_RobotPose2DIIS.data.heading          = Status.odotheta;

    //outodo
    //printf("odo %f %f %f\n", m_RobotPosIIS.x, m_RobotPosIIS.y, m_RobotPosIIS.theta);
  }




  RTC::Time tmodo;
  tmodo.sec  = (::CORBA::ULong) Status.odotime.sec();
  tmodo.nsec = (::CORBA::ULong) (Status.odotime.usec() * 1000);

  m_EncoderL.tm     = tmodo;
  m_EncoderL.data   = encoder[0];
  m_EncoderR.tm     = tmodo;
  m_EncoderR.data   = encoder[1];

  m_Odometry.tm     = tmodo;
  m_Odometry.x      = Status.odox;
  m_Odometry.y      = Status.odoy;
  m_Odometry.theta  = Status.odotheta;

  m_OdometryB.tm    = tmodo;
  m_OdometryB.x     = m_RobotPosIIS.x;
  m_OdometryB.y     = m_RobotPosIIS.y;
  m_OdometryB.theta = m_RobotPosIIS.theta;


  m_EncoderLOut.write();
  m_EncoderROut.write();
  m_OdometryOut.write();
  m_OdometryBOut.write();
  

  //IIS
  m_RobotVelIISOut.write();
  m_RobotVel2DIISOut.write();
  
  m_RobotPosIISOut.write();
  m_RobotPose2DIISOut.write();

  //usleep(1000);

  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t Blackship::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Blackship::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Blackship::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Blackship::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Blackship::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{

  void BlackshipInit(RTC::Manager* manager)
  {
    RTC::Properties profile(blackship_spec);
    manager->registerFactory(profile,
                             RTC::Create<Blackship>,
                             RTC::Delete<Blackship>);
  }

};


